/* 
   COPYRIGHT 2006 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - socklib.c */
/* rdk@krupczak.org */

/* a basic set of socket library routines attempting to isolate
   the messiness of BSD sockets; note that we do not do anything
   as of yet with IPv6; everything here assumes IPv4 XXX; 
   on Linux, everything needs to be in network-byte order a prior 
   see "man 7 ip" ; most functions appear to return things
   in network byte order as well

   support, on non-win32, for ipv6 address conversions added
   but not connecting/talking over ipv6 yet (3/21/2006)
*/


#ifdef WIN32
#include <windows.h>
#include <IPExport.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <strings.h>
#endif

#ifdef SOLARIS
#include <sys/filio.h> /* for FIONREAD */
#endif

#ifndef WIN32
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#endif

#ifdef SOLARIS
#include <arpa/inet.h>
#include <inttypes.h>
#endif

#include "socklib.h"

/* ****************************************************************** */
/* get a UDP/Internet socket and return it */

/* fortunately, in winsock, a SOCKET type is really an u_int */

int get_udp_socket()
{
     return socket(PF_INET,SOCK_DGRAM,IPPROTO_UDP);
}

/* ****************************************************************** */
/* get a TCP/Internet socket and return it */

int get_tcp_socket()
{
     return socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
}

/* ****************************************************************** */
/* bind a UDP socket to the specified port (network byte order) */
/* if reuseaddr=1, setsockopt to SO_REUSEADDR */

int bind_udp_socket(int s, u_int ipaddr, u_short port, int reuseaddr)
{
     struct sockaddr_in addr;
     int flag;

     if (reuseaddr) {
        flag = 1; 
        setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&flag,sizeof(flag));
     }

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return bind(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* bind a TCP/Internet socket to the specified port (network byte order) */
/* if reuseaddr=1, setsockopt to SO_REUSEADDR */

int bind_tcp_socket(int s, u_int ipaddr, u_short port, int reuseaddr)
{
     struct sockaddr_in addr;
     int flag;

     if (reuseaddr) {
        flag = 1; 
        if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&flag,
                       sizeof(flag)) <0) {
 	   fprintf(stderr,"bind_tcp_socket: setsockop SO_REUSEADDR failed, %d",
                   errno);
        }
        else {
	   ; 
        }
     }

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return bind(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* connect a TCP socket to dest addr, dest port (net byte order) */

int connect_tcp_socket(int s, u_int ipaddr, u_short port)
{
     struct sockaddr_in addr;

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return connect(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* connect a TCP socket to dest addr/port (net byte order) waiting
   at most timeout seconds; if timeout or no connection, return -1; */

int connect_tcp_socket_timeout(int s, u_int ipaddr, u_short port, int timeout)
{
     struct sockaddr_in addr;
     struct sockaddr peeraddr;
     fd_set write_set;
     fd_set error_set;
     struct timeval timeoutval;
     int ret,on,peernamelen;

     /* set socket non-blocking */
     on = 1;
#ifndef WIN32
     ioctl(s,FIONBIO,&on);
#else
     ioctlsocket(s,FIONBIO,&on);
#endif

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     ret = connect(s,(struct sockaddr *)&addr,sizeof(addr));

     /* DBG2("connect_tcp_socket_timeout: initial connect ret %d, errno:%d\n",ret,errno);*/

     if (ret == 0) {
       /* DBG("connect_tcp_socket_timeout: immediate connection\n");*/
        /* immediate connection */
        return 1;
     }

     /* on solaris, this condition seems to occur; dont panic, 
        it appears "normal" proceed with async connection */
     if ((ret != 0) && (errno == 0)) {
        ; /* continue w/o error */
     }

#ifndef WIN32
     /* check for error and errno=EINPROGRESS */
     if ((errno != 0) && (errno != EINPROGRESS)) {
       /* DBG("connect_tcp_socket_maxdelay: error %d\n",errno);*/
        return -1;
     }
#else
     /* check for SOCK_ERROR and WSAGetLastError() = WSAEWOULDBLOCK */
#endif

     /* select for up to max seconds; if connection is available,
        then it will be available for writing */

     FD_ZERO(&write_set);
     FD_SET(s,&write_set);
     FD_ZERO(&error_set);
     FD_SET(s,&error_set);

     timeoutval.tv_sec = timeout;
     timeoutval.tv_usec = 0;
     
     if ((ret = select(s+1,NULL,&write_set,&error_set,&timeoutval)) <= 0) {
        /* timeout, connection not completed, nothing pending */
       /* DBG("connect_tcp_socket_timeout: timeout or error\n");*/
        return -1;
     }

     if (FD_ISSET(s,&write_set) == 0) {
       /* DBG("connect_tcp_socket_timeout: no connection\n"); */
        return -1;
     }

     if (FD_ISSET(s,&error_set) != 0) {
       /* DBG1("connect_tcp_socket_timeout: error on socket %d\n",s);*/
        return -1;
     }

     /* in some cases (Solaris, Linux), we get this far but we really
        havent connected and we then generate SIGPIPE later on; test
        if connection succeeded by calling getpeername() */
     peernamelen = sizeof(peeraddr);
     if (getpeername(s,&peeraddr,&peernamelen) < 0) {
       /* DBG1("connect_tcp_socket_timeout: getpeername failed, %d\n",errno);*/
        return -1;
     }

     /* DBG("connect_tcp_socket_timeout: seems to have connected\n"); */

     /* set back to blocking? Yes, since that is the semantics
        that the rest of our socklib assumes */
     on = 0;
#ifndef WIN32
     ioctl(s,FIONBIO,&on);
#else
     ioctlsocket(s,FIONBIO,&on);
#endif

     return 1;
}

/* ****************************************************************** */
/* connect UDP/Internet socket to destaddr, destport */

int connect_udp_socket(int s, u_int ipaddr, u_short port)
{
     struct sockaddr_in addr;

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return connect(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* listen for a TCP connection request from some client */

int listen_for_tcp_connection(int s, int backlog)
{
     return listen(s,backlog);
}

/* ****************************************************************** */
/* check to see if there is a connection attempt pending
   on socket s that is in listen state; wait for up to t seconds
   before returning; return 0 or 1 */

int connection_pending(int s, int timeout)
{
     fd_set fdset;
     struct timeval timeoutval;
     int ret;

     FD_ZERO(&fdset);
     FD_SET(s,&fdset);
     timeoutval.tv_sec = timeout;
     timeoutval.tv_usec = 0;

     ret = select(s+1,&fdset,NULL,NULL,&timeoutval);
     if (ret == 0) {
        /* timeout, nothing interesting, no connections pending */
        return 0;
     }
 
     /* we could check via FD_ISSET */

     return 1;
}

/* ****************************************************************** */
/* accept a TCP connection request and return new socket as return arg */
/* this function blocks until connection request */

int accept_tcp_connection_request(int s, u_int *ipaddr, u_short *port)
{
     struct sockaddr_in addr;
     int ret,namelen;

     if ((ipaddr == NULL) || (port == NULL)) {
        fprintf(stderr,"accept_tcp_connection: invalid argument\n");
        return -1;
     }

     namelen = sizeof(addr);

     ret = accept(s,(struct sockaddr *)&addr,&namelen);

     *ipaddr = addr.sin_addr.s_addr;
     *port = addr.sin_port;

     return ret;
}

/* ****************************************************************** */
/* given name, convert to ipV4addr -- grab first address in list */
/* in network byte order */

u_int hostname_to_ipaddr(char *hostname)
{
     struct hostent *h;
     u_int addr;

     if (hostname == NULL) {
        return (u_int)0;
     }

     addr = inet_addr(hostname);
     if (addr == (u_int)-1) {

        if ((h = gethostbyname(hostname)) == NULL) {
#if 0
#ifdef LINUX
            herror("gethostbyname failed");
#else
            perror("gethostbyname failed");
#endif
           fprintf(stderr,". . . .for %s\n",hostname);
#endif /* 0 */
           return (u_int)0;
        }

        if (h->h_length != 4) {
           fprintf(stderr,"hostname_to_ipaddr: address of length %d\n",
                   h->h_length);
           return 0;
        }
        if (h->h_addr == NULL) {
 	   fprintf(stderr,"hostname_to_ipaddr: h_addr is null\n");
	   return 0;
        }

        addr = *((u_int *)(h->h_addr));

     } /* inet_addr == -1 */

     return addr;
}

/* ****************************************************************** */
/* given an IP address (in network byte order), convert and
   print via the inet_ntoa */

int convert_ipaddr(u_int ipaddr, char *ipaddrstr, int maxlen)
{
     char *str;
     struct in_addr in;

     if ((ipaddrstr == NULL) || (maxlen < 8)) {
        return -1;
     }

     in.s_addr = ipaddr;
     str = inet_ntoa(in);

     strncpy(ipaddrstr,str,maxlen-1);

     return 1;
}

int convert_ipv6addr(u_char *ipv6addr, char *ipv6str, int maxlen)
{
     struct in6_addr *addr6;

     if ((ipv6str == NULL) || (maxlen < 32)) {
        return -1;
     }

     ipv6str[0] = 0;

     addr6 = (struct in6_addr *)ipv6addr;

     /* take 4-char chunks at a time as per RFC 2373 */
     sprintf(ipv6str,"%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
             addr6->s6_addr[0],
             addr6->s6_addr[1],
	     addr6->s6_addr[2],
	     addr6->s6_addr[3],
	     addr6->s6_addr[4],
	     addr6->s6_addr[5],
	     addr6->s6_addr[6],
	     addr6->s6_addr[7],
	     addr6->s6_addr[8],
	     addr6->s6_addr[9],
	     addr6->s6_addr[10],
	     addr6->s6_addr[11],
	     addr6->s6_addr[12],
	     addr6->s6_addr[13],
	     addr6->s6_addr[14],
	     addr6->s6_addr[15]);

     return 1;
}

/* ****************************************************************** */
/* 
   keep reading, from socket s, into buf[maxlen] until X bytes are 
   received or until some recv error
*/
  
int read_upto_X_bytes(int s, u_char *buf, int X, int maxlen)
{
     int retlen = 0;
     int ret;

     if (buf == NULL) {
        return -1;
     }
     if (maxlen < X) {
        fprintf(stderr,"read_upto_X_bytes: not enough space, %d < %d\n",
                maxlen,X);
        return -1;
     }

     while ((X - retlen) > 0) {

       if ((ret = recv(s,(void *)&buf[retlen],X-retlen,0)) < 0)
 	  break;

       retlen += ret;
     }

     /* printf("read_upto_X_bytes: read %d bytes\n",retlen); */

     return retlen;
}

/* ****************************************************************** */
/* check to see if data at socket to read, and how much */

#ifndef WIN32

/* UNIX */

int bytes_at_socket(int s)
{
     int bytes;

     if (s == 0) {
        return 0;
     }

     bytes = 0;
     if (ioctl(s,FIONREAD,(char *)&bytes) < 0)
        perror("ioctl FIONREAD failed");

     return bytes;
}

#else /* WIN32 */

int bytes_at_socket(int s)
{
     u_long bytes;

     if (s == 0) {
        return 0;
     }

     bytes = 0;

     if (ioctlsocket(s,FIONREAD,(u_long *)&bytes) != 0)
        perror("ioctl FIONREAD failed");

     return (int)bytes;
}

#endif /* !WIN32 */

/* ****************************************************************** */
/* given a TCP socket, turn off buffering so that all bytes get sent
   when we call send -- however, setting this still does not guarantee 
   that TCP can deliver on message boundaries as TCP fundamentally is 
   a byte-stream interface, not a message interface 
*/

int set_tcp_no_delay(int s)
{
     int ret;
     int nodelay = 1;

     if ((ret = setsockopt(s,IPPROTO_TCP,TCP_NODELAY,(char *)&nodelay,
                           sizeof(nodelay))) < 0)
        perror("setsockopt: tcp_nodelay failed");

     return ret;
 
}

/* ****************************************************************** */
/* set a socket's send/receive buffer sizes -- UDP or TCP */
/* if either size is 0, dont set that buffer's size */

int set_socket_buffer_sizes(int s, int recvsize, int sendsize)
{
     if (recvsize > 0) {
        if (setsockopt(s,SOL_SOCKET,SO_RCVBUF,(char *)&recvsize,sizeof(int))<0)
	   perror("setsockopt of SO_RCVBUF failed");
     }

     if (sendsize > 0) {
        if (setsockopt(s,SOL_SOCKET,SO_SNDBUF,(char *)&sendsize,sizeof(int))<0)
	   perror("setsockopt of SO_SNDBUF failed");
     }

     return 1;
}

/* ****************************************************************** */
/* get socket buffer send/receive size -- UDP or TCP socket */
/* if null ptr for either, dont retrieve/return that one */

int get_socket_buffer_sizes(int s, int *recvsize, int *sendsize) 
{
     int sndbufsz, rcvbufsz;
     int optlen;

     if (recvsize != NULL) {
        *recvsize = -1;
        optlen = sizeof(int);
        if (getsockopt(s,SOL_SOCKET,SO_RCVBUF,(char *)&rcvbufsz,&optlen) >= 0)
 	   *recvsize = rcvbufsz;
        else
	   perror("getsockopt of SO_RCVBUF failed");
     } 

     if (sendsize != NULL) {
        *sendsize = -1;
        optlen = sizeof(int);
        if (getsockopt(s,SOL_SOCKET,SO_SNDBUF,(char *)&sndbufsz,&optlen) >= 0)
 	   *sendsize = sndbufsz;
        else
	   perror("getsockopt of SO_SNDBUF failed");
     } 

     return 1;
}

/* ****************************************************************** */
/* given a address/string (hostname or IP address), convert it
   to a (hopefully FQDN) hostname of max lenght maxlen and return */

int addressToHostname(char *addrstr, char *hostname, int maxlen)
{
     struct in_addr a;
     struct hostent *h;

     if ((addrstr == NULL) || (hostname == NULL) || (maxlen < 5))
        return -1;

     /* default, in case of error, is simply to return what 
        was passed in 
      */
     strncpy(hostname,addrstr,maxlen);
     hostname[maxlen-1] = 0;

     /* get matching address; if its a hostname,
        0 returned and we try to get address 
      */ 
#if defined(WIN32) || defined(SOLARIS)
     /* inet_aton() not supported on win32 and Solaris */
#ifdef SOLARIS
#define INADDR_NONE -1
#endif

     if ((a.s_addr = inet_addr(addrstr)) == INADDR_NONE) {

        if ((h = gethostbyname(addrstr)) == NULL) {
           return 1;
        }
        /* check to make sure h->h_addr is valid */
        if (h->h_addr) {
           a.s_addr = *((u_int *)(h->h_addr));
        }
        else {
	   return -1;
        }
     }
#else
     if (inet_aton(addrstr,&a) == 0) {

        if ((h = gethostbyname(addrstr)) == NULL) {
           return 1;
        }
        /* check to make sure h->h_addr is valid */
        if (h->h_addr) {
           a.s_addr = *((u_int *)(h->h_addr));
        }
        else {
	   return -1;
        }
     }
#endif

     if ((h = gethostbyaddr((char *)&a,sizeof(a),AF_INET)) == NULL) {
        return 1;
     }
     if (h->h_name == NULL) {
        return 1;
     }

     strncpy(hostname,h->h_name,maxlen);
     hostname[maxlen-1] = 0;

     return 1;
}

/* ****************************************************************** */
