/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmp.c */
/* rdk@krupczak.org */

/* 
   XML message parsing functions, etc.
   These routines are what we call to convert between
   XML documents and internal PDU formats.
   We use libxml2 http://www.xmlsoft.org/ 
   as our xml parser for all platforms.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <unistd.h>
#include <strings.h>
#endif

#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "xmp.h"

extern int xmp_session_verbose;

/* ************************************************************* */
/* given an internal-format message, build an XML-encoded message 
   and place it in buffer which is pre-allocated
   We're assuming the PDU is correctly formed

   We use sprintf() and step through the provided buffer
   using return from sprintf to step through buffer; only
   one zero-termination at end of entire message and 
   that is not included in overall length

   Return length of message (w/o terminating zero)

   if we run out of room, stop, return -1 and set needed
   to the amount of space in bytes that we think we will need
*/

int buildXmpMessage(xmpPdu_t *pdu, u_char *buffer, int buflen, int *needed)
{
     xmpVariable_t *varptr;
     int len, ret, estimate;
     char aStr[256];
     int x;

     if ((pdu == NULL) || (buffer == NULL)) {
        fprintf(stderr,"buildXmpMessage: invalid args\n");
        return -1;
     }

     if ((u_char *)pdu == buffer) {
        fprintf(stderr,"buildXmpMessage: pdu = buffer!\n");
     }

     /* make guesstimate of length of XML messages */
     estimate = 355                    /* length of template */
          + 140*pdu->numberVariables;  /* length of variable/templates+\n's */
     varptr = pdu->varList;
     x = 0;
     while (varptr != NULL) {
       estimate += varptr->valueStrLen + 2; /* value, quotes */
       estimate += 20;  /* max length of type-string */
       estimate += strlen(varptr->varName.mibName) +
	           strlen(varptr->varName.objName) +
             	   strlen(varptr->varName.key) + 5;
       x++;
       varptr = varptr->next;
     }

     if (x != pdu->numberVariables) {
        fprintf(stderr,"buildXmpMessage: numVariables does not match %d %d\n",
                x,pdu->numberVariables);
     }

     if (estimate > buflen) {
        if (xmp_session_verbose)
           fprintf(stderr,"buildXmpMessage: insufficient buffer size, %d %d\n",
                   estimate,buflen);
        /* need to return an error */
        if (needed != NULL)
	   *needed = estimate;
	return -1;
     }

     /* start counting  . . . */
     len = 0;

     /* make xml header */
     ret = sprintf((char *)&buffer[len],
                   "<?xml version=\"1.0\" encoding='US-ASCII' ?>\n");
     len += ret;
     ret = sprintf((char *)&buffer[len],"<!DOCTYPE xmp>\n");
     len += ret;
 
     /* make xml message type, messageID */
     ret = sprintf((char *)&buffer[len],"<xmp message-id=\"%lld\">\n",
                   pdu->messageID);
     len += ret;

     DBG1("buildXmpMessage: message-id is %lld\n",pdu->messageID);

     /* make XMP message type */
     messageTypeToString(pdu->msgType,aStr,sizeof(aStr));
     ret = sprintf((char *)&buffer[len],"<%s>\n",aStr);
     len += ret;

     /* error status/index */
     errorStatusToString(pdu->errorStatus,aStr,sizeof(aStr));
     ret = sprintf((char *)&buffer[len],"<error-status>%s</error-status>\n",
                   aStr);
     len += ret;
     ret = sprintf((char *)&buffer[len],"<error-index>%d</error-index>\n",
                   pdu->errorIndex);
     len += ret;

     /* if its a table operation, put in <tableInfo> */
     if ((pdu->msgType > SetRequest) && (pdu->msgType < Trap)) {
        ret = sprintf((char *)&buffer[len],"<tableInfo>\n");
        len += ret;

        ret = sprintf((char *)&buffer[len],"<mib>");
        len += ret;
        ret = sprintf((char *)&buffer[len],"%s",pdu->tableMib);
        len += ret;
        ret = sprintf((char *)&buffer[len],"</mib>\n");
        len += ret;
        
        ret = sprintf((char *)&buffer[len],"<object>");
        len += ret;
        ret = sprintf((char *)&buffer[len],"%s",pdu->tableName);
        len += ret;
        ret = sprintf((char *)&buffer[len],"</object>\n");
        len += ret;

        ret = sprintf((char *)&buffer[len],"<key>");
        len += ret;
        ret = sprintf((char *)&buffer[len],"%s",pdu->tableKey);
        len += ret;
        ret = sprintf((char *)&buffer[len],"</key>\n");
        len += ret;

        ret = sprintf((char *)&buffer[len],"<max-rows>");
        len += ret;
        ret = sprintf((char *)&buffer[len],"%d",pdu->maxRows);
        len += ret;
        ret = sprintf((char *)&buffer[len],"</max-rows>\n");
        len += ret;

        ret = sprintf((char *)&buffer[len],"</tableInfo>\n");
        len += ret;
     }

     /* variable list */

     ret = sprintf((char *)&buffer[len],"<variableList>\n");
     len += ret;

     /* the variables */

     /* we should periodically compare len and buflen and make sure
        we stop before overflow and return needed;
        since we're now doing a much better job of estimating packet
        size, we're probably safe
     */

     varptr = pdu->varList;
     while (varptr != NULL) {
       ret = sprintf((char *)&buffer[len],"<variable>\n");
       len += ret;

       /* varname */
       if (varptr->varName.key[0] == 0) {
 	  /* no key, tag with no content */
          ret = sprintf((char *)&buffer[len],
                "<varname><mib>%s</mib><object>%s</object><key/></varname>\n",
		     varptr->varName.mibName,
		     varptr->varName.objName);
       }
       else {
 	  /* key, tag with content */
          ret = sprintf((char *)&buffer[len],
          "<varname><mib>%s</mib><object>%s</object><key>%s</key></varname>\n",
		     varptr->varName.mibName,
		     varptr->varName.objName,
		     varptr->varName.key);
       }
       len += ret;

       /* value/valuetype */
       if (varptr->varSyntax == nullSyntax) {
          ret = sprintf((char *)&buffer[len],
                        "<varvalue vartype=\"nullSyntax\"/>\n");
       }
       else {
          syntaxToString(varptr->varSyntax,aStr,sizeof(aStr));

          if (varptr->valueBuf != NULL) {
             ret = sprintf((char *)&buffer[len],
                        "<varvalue vartype=\"%s\">%s</varvalue>\n",
                        aStr,varptr->valueBuf);

          }
          else {
	     /* assumes zero termination!! */
             ret = sprintf((char *)&buffer[len],
                        "<varvalue vartype=\"%s\">%s</varvalue>\n",
                        aStr,varptr->valueStr);
          }

       }
       len += ret;

       /* close variable tag */
       ret = sprintf((char *)&buffer[len],"</variable>\n");
       len += ret;

       /* next variable */
       varptr = varptr->next;

     } /* while variables in list */

     /* close tags */
     ret = sprintf((char *)&buffer[len],"</variableList>\n");
     len += ret;
     messageTypeToString(pdu->msgType,aStr,sizeof(aStr));
     ret = sprintf((char *)&buffer[len],"</%s>\n",aStr);
     len += ret;
     ret = sprintf((char *)&buffer[len],"</xmp>\n");
     len += ret;

     if (len > buflen) {
       fprintf(stderr,
              "buildXmpMessage: overwritten mem buf, %d>%d, est:%d, vars:%d\n",
               len,buflen,estimate,pdu->numberVariables);
     }

     if (xmp_session_verbose)
        fprintf(stderr,"buildXmpMessage: len:%d, buflen:%d, estimate:%d\n",
                len,buflen,estimate);
  
     return len;
}

/* ************************************************************* */
/* copy an internal PDU from src to dst; variables are malloc'd
   and copied, not referenced
 */

int copyXmpPdu(xmpPdu_t *src, xmpPdu_t *dst)
{
     xmpVariable_t *sptr, *dptr;

     if ((src == NULL) || (dst == NULL)) {
        fprintf(stderr,"copyXmpPdu: invalid args\n");
        return -1;
     }

     /* this copies everything, including pointers */
     bcopy((void *)src,(void *)dst,sizeof(xmpPdu_t));

     /* now, go and actually malloc/copy the variables */
     dst->varList = NULL;
     dst->varListTail = NULL;
     sptr = src->varList;

     while (sptr != NULL) {
       dptr = copyXmpVariable(sptr); /* XXX check for error */
       dptr->next = NULL;
 
       if (dst->varList == NULL)
	  dst->varList = dptr;
       if (dst->varListTail == NULL) {
 	  dst->varListTail = dptr;
       }
       else {
 	  dst->varListTail->next = dptr;
          dst->varListTail = dptr;          
       }
       
       sptr = sptr->next;
     }

     return 1;
}

/* ************************************************************* */
/* copy an XmpVariable including valueBuf if present */

xmpVariable_t *copyXmpVariable(xmpVariable_t *v)
{
     xmpVariable_t *copy;

     if (v == NULL)
        return NULL;

     copy = (xmpVariable_t *)malloc(sizeof(xmpVariable_t));
     if (copy != NULL) {
        bcopy(v,copy,sizeof(xmpVariable_t));

        if (v->valueBuf != NULL) {
           copy->valueBuf = (u_char *)malloc(v->valueStrLen+1);
           if (copy->valueBuf != NULL)
              bcopy(v->valueBuf,copy->valueBuf,v->valueStrLen+1);
        }

        copy->next = NULL;
     }

     return copy;
}

/* ************************************************************* */
/* given an internal-format PDU, clean it up by
   free'ing each and any xmpVariable_t that may be on variable list 
   separately; this assumes that each variable was malloc'd 
   separately and is not part of a larger block!!
 
   We do not free the PDU structure itself though.

*/

int xmpPduCleanup(xmpPdu_t *pdu)
{
     xmpVariable_t *ptr;
     int i = 0;
  
     if (pdu == NULL) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpPduCleanup: null ptr, no cleanup\n");
        return -1;
     }

     if (pdu->numberVariables == 0) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpPduCleanup: numvars is 0, no cleanup\n");
        return 0;
     }
     
     if (pdu->varList == NULL) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpPduCleanup: no varlist, no cleanup\n");
        return 0;
     }

     DBG1("xmpPduCleanup: %d vars\n",pdu->numberVariables);
     /* dumpInternalXmpMessage(pdu); */

     ptr = pdu->varList;
     while (ptr != NULL) {
       pdu->varList = ptr->next;

       if (ptr->valueBuf != NULL) {
	  DBG1("xmpPduCleanup: freeing valueBuf %lx\n",ptr->valueBuf);
	  free(ptr->valueBuf);
       }  

       free(ptr);
       i++;
       
       ptr = pdu->varList;
     }
     
     if (i != pdu->numberVariables) {
        fprintf(stderr,"xmpPduCleanup: i=%d, numberVariables=%d\n",
   	       i,pdu->numberVariables);
     }
     /* assert(i == pdu->numberVariables); */
     pdu->varList = pdu->varListTail = NULL;

     DBG("xmpPduClenaup: done\n");

     return i;
}

/* ************************************************************* */
/* parse XMP table info -- 
   <tableInfo>
    <mib></mib>
    <object></object>
    <key></key>
    <max-rows></max-rows>
   </tableInfo>
*/

int parseXmpTableInfo(xmlNodePtr node, xmpPdu_t *pdu)
{
     xmlNodePtr curNode;

     if ((node == NULL) || (pdu == NULL)) {
        fprintf(stderr,"parseXmpTableInfo: invalid arg(s)\n");
        return -1;
     }

     if (strcmp((char *)node->name,"tableInfo") != 0) {
        fprintf(stderr,"%s is not tableInfo\n",node->name);
        return -1;
     }

     if (node->children == NULL) {
        fprintf(stderr,"tableInfo does not have mib, object, key, etc.\n");
        return -1;
     }

     /* <mib></mib> */
     curNode = node->children;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"tableInfo does not contain mib . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"mib") != 0) {
        fprintf(stderr,"'%s' is not mib\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No mib value\n");
        return -1;
     }
     strncpy(pdu->tableMib,(char *)curNode->children->content,127);
     pdu->tableMib[127] = 0;

     /* <object></object> */
     curNode = curNode->next;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"tableInfo does not contain object . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"object") != 0) {
        fprintf(stderr,"'%s' is not table object\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No table object value\n");
        return -1;
     }
     strncpy(pdu->tableName,(char *)curNode->children->content,127);
     pdu->tableName[127] = 0;

     /* <key></key> */
     curNode = curNode->next;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"tableInfo does not contain table key . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"key") != 0) {
        fprintf(stderr,"'%s' is not table key\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No table key value\n");
        return -1;
     }
     strncpy(pdu->tableKey,(char *)curNode->children->content,2047);
     pdu->tableName[2047] = 0;

     /* <max-rows></max-rows> */
     curNode = curNode->next;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"tableInfo does not contain max-rows . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"max-rows") != 0) {
        fprintf(stderr,"'%s' is not table max-rows\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No table max-rows value\n");
        return -1;
     }
     pdu->maxRows = atoi((char *)curNode->children->content);

     return 1;     
}

/* ************************************************************* */
/* parse an XMP variable; malloc space for variable 
   and put it on variableList 
*/

int parseXmpVariable(xmlNodePtr varNode, xmpPdu_t *pdu)
{
     xmlNodePtr curNode, nodePtr;
     xmpVariable_t *varPtr;
     int valueLen;

     DBG("parseXmpVariable: starting\n");

     if ((varNode == NULL) || (pdu == NULL)) {
        fprintf(stderr,"parseXmpVariable: invalid arg(s)\n");
        return -1;
     }

     if (strcmp((char *)varNode->name,"variable") != 0) {
        fprintf(stderr,"%s is not a variable\n",varNode->name);
        return -1;
     }

     if ((varPtr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        perror("Malloc of variable failed");
        return -1;
     }
     varPtr->valueBuf = NULL; /* default to storing value internal */

     DBG("parseXmpVariable: going for varname\n");

     /* varname, varvalue are children of variable */
     if (varNode->children == NULL) {
        fprintf(stderr,"variable does not have name,value\n");
        free(varPtr);
        return -1;
     }

     /* parse varname -- mib, object, and key should be children
        and we use curNode and nodePtr for parsing them 
      */

     /* parse varname ******************* */
     curNode = varNode->children;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"No varname\n");
        free(varPtr);
        return -1;
     }
     if (strcmp((char *)curNode->name,"varname") != 0) {
        fprintf(stderr,"%s should be varname\n",curNode->name);
        free(varPtr);
        return -1;
     }
     nodePtr = curNode->children;
     while (xmlIsBlankNode(nodePtr)) { nodePtr = nodePtr->next; }
     if (nodePtr == NULL) {
        fprintf(stderr,"No mib/object tags\n");
        free(varPtr);
        return -1;
     }
     if (strcmp((char *)nodePtr->name,"mib") != 0) {
        fprintf(stderr,"%s should be mib\n",nodePtr->name);
        free(varPtr);
        return -1;
     }
     if (nodePtr->children == NULL) {
        fprintf(stderr,"No mib value\n");
        free(varPtr);
        return -1;
     }
     /* get content of mib tag */
     strncpy(varPtr->varName.mibName,(char *)nodePtr->children->content,127);
     varPtr->varName.mibName[127] = 0;

     DBG("parseXmpVariable: got mib name\n");

     nodePtr = nodePtr->next;
     while (xmlIsBlankNode(nodePtr)) { nodePtr = nodePtr->next; }
     if (nodePtr == NULL) { 
        fprintf(stderr,"missing object tag\n");
        free(varPtr);
        return -1;
     }
     if (strcmp((char *)nodePtr->name,"object") != 0) {
        fprintf(stderr,"%s should be object\n",nodePtr->name);
        free(varPtr);
        return -1;
     }
     if (nodePtr->children == NULL) {
        fprintf(stderr,"No object value\n");
        free(varPtr);
        return -1;
     }
     /* get content of object tag */
     strncpy(varPtr->varName.objName,(char *)nodePtr->children->content,127);
     varPtr->varName.objName[127] = 0;

     DBG("parseXmpVariable: got object\n");

     /* key, key-value, etc. */
     varPtr->varName.key[0] = 0; /* default */
     nodePtr = nodePtr->next;
     while (xmlIsBlankNode(nodePtr)) { nodePtr = nodePtr->next; }
     if (nodePtr == NULL) { 
        fprintf(stderr,"missing key tag\n");
        free(varPtr);
        return -1;
     }
     if (strcmp((char *)nodePtr->name,"key") != 0) {
        fprintf(stderr,"%s should be key\n",nodePtr->name);
        free(varPtr);
        return -1;
     }
     /* key value is optional */
     if (nodePtr->children != NULL) {
        /* get content of object tag */
        strncpy(varPtr->varName.key,(char *)nodePtr->children->content,2047);
        varPtr->varName.key[2047] = 0;
     }

     /* parse varvalue */
     curNode = curNode->next;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        /* dont forget to check curNode; thanks to jeffg for finding
           this omission */
        fprintf(stderr,"missing varvalue tag\n");
        free(varPtr);
        return -1;
     }
     if (strcmp((char *)curNode->name,"varvalue") != 0) {
        fprintf(stderr,"%s should be varvalue\n",curNode->name);
        free(varPtr);
        return -1;
     }

     /* vartype is attribute of varvalue, must be present */
     if (curNode->properties == NULL) {
        fprintf(stderr,"No vartype present\n");
        free(varPtr);
        return -1;
     }
     if (strcmp((char *)curNode->properties->name,"vartype") != 0) {
        fprintf(stderr,"%s should be vartype\n",curNode->properties->name);
        free(varPtr);
        return -1;
     }
     if (curNode->properties->children == NULL) {
        fprintf(stderr,"vartype has no value\n");
        free(varPtr);
        return -1;
     }
     varPtr->varSyntax = stringToSyntax(
                         (char *)curNode->properties->children->content);

     DBG("parseXmpVariable: got var type\n");
     
     /* the varvalue itself (if present) would be child node */
     nodePtr = curNode->children;
     varPtr->valueStr[0] = 0;
     varPtr->valueStrLen = 0;
     while (xmlIsBlankNode(nodePtr)) { nodePtr = nodePtr->next; }     
     if (nodePtr != NULL) {
        /* printf("Valuestr '%s'\n",nodePtr->content); */
        valueLen = strlen((char *)nodePtr->content);
        if (valueLen > 4095) {
	   DBG1("parseXmpVariable: value len %d > 4095\n",valueLen);
           varPtr->valueStrLen = 0;
           varPtr->valueBuf = (u_char *)malloc(valueLen+2); /* extra for \0 */
           if (varPtr->valueBuf != NULL) {
              varPtr->valueStrLen = valueLen;
	      strncpy((char *)varPtr->valueBuf,(char *)nodePtr->content,
                      valueLen);
              varPtr->valueBuf[varPtr->valueStrLen] = 0;
           }
        }
        else {
	   varPtr->valueBuf = NULL; /* just to make sure */
           strncpy(varPtr->valueStr,(char *)nodePtr->content,4095);
           varPtr->valueStr[4095] = 0;
           varPtr->valueStrLen = strlen(varPtr->valueStr);
        }
     }

     /* put variable onto tail of variable list, increment
        number of variables */
     pdu->numberVariables++;
     if (pdu->varListTail != NULL)
        pdu->varListTail->next = varPtr;
     varPtr->next = NULL;
     pdu->varListTail = varPtr;
     if (pdu->varList == NULL)
        pdu->varList = varPtr;

     return 1;
}


/* ************************************************************* */
/* parse an XMP message stored in file */

int parseXmpMessageByFilename(char *filename, xmpPdu_t *pdu)
{
     FILE *fp;
     u_char *buffer;
     int bufsize,ret;
     struct stat statbuf;

     if ((filename == NULL) || (pdu == NULL)) {
        fprintf(stderr,"parseXmpMessageByFilename: invalid args\n");
        return -1;
     }
     
     /* stat the file to see how much space to malloc */
#ifdef WIN32
     if (_stat(filename,&statbuf) < 0) {
#else
     if (stat(filename,&statbuf) < 0) {
#endif
        perror("stat failed");
        return -1;
     }
     
     /* if its a regular file */
#ifdef WIN32
#define	_S_ISTYPE(mode, mask)	(((mode) & _S_IFMT) == (mask))
#define	S_ISREG(mode)	 _S_ISTYPE((mode), _S_IFREG)
#endif

     if (! S_ISREG(statbuf.st_mode)) {
        fprintf(stderr,"%s is not regular\n",filename);
        return -1;
     }

     /* malloc the space needed */
     if ((buffer = (u_char *)malloc(statbuf.st_size + 5)) == NULL) {
        perror("malloc failed");
        return -1;
     }

     bufsize = statbuf.st_size + 5;

     if ((fp = fopen(filename,"r")) == NULL) {
        perror("fopen file failed");
        free(buffer);
        return -1;
     }

     ret = fread(buffer,1,statbuf.st_size,fp);
     if (ret != statbuf.st_size) {
        fprintf(stderr,"Read %d rather than %ld\n",ret,statbuf.st_size);
        free(buffer);
        fclose(fp);
        return -1;
     }

     fclose(fp);

     ret = parseXmpMessage(buffer,statbuf.st_size,pdu);

     free(buffer);

     /* clean up the PDU?  Caller needs to do this */
     
     return ret;
}

/* ************************************************************* */
/* 
   parse an XMP message, stored in memory buffer, into internal
   format; on error, we call cleanup which frees variables
   that were malloc'd 
*/

int parseXmpMessage(u_char *membuf, int bufsize, xmpPdu_t *pdu)
{
     xmlDocPtr doc;
     xmlNodePtr cur;

     if ((membuf == NULL) || (pdu == NULL)) {
        fprintf(stderr,"parseXmpMessage: invalid arg(s) membuf/pdu\n");
        return -1;
     }
     if (bufsize < 25) {
        fprintf(stderr,"parseXmpMessage: invalid bufsize, %d\n",bufsize);
        return -1;
     }

     if ((doc = xmlParseMemory((char *)membuf,bufsize)) == NULL) {
        fprintf(stderr,"parseXmpMessage: xml parse error\n");
        return -1;
     }

     /* get doc root node; it should be <xmp> */

     if ((cur = xmlDocGetRootElement(doc)) == NULL) {
        fprintf(stderr,"parseXmpMessage: empty document\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->name,"xmp") != 0) {
        fprintf(stderr,"parseXmpMessage: document is not an XMP message, %s\n",
                cur->name);
        xmlFreeDoc(doc);
        return -1;
     }
     
     /* get message-id as attribute  of XMP tag */
     if ((cur->properties == NULL) || 
         (cur->properties->children == NULL)) {
        fprintf(stderr,"parseXmpMessage: XMP message has no id\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->properties->name,"message-id") != 0) {
        fprintf(stderr,"parseXmpMessage: XMP message has no id\n");
        xmlFreeDoc(doc);
        return -1;
     }

     DBG1("parseXmpMessage: xml message-ID is '%s'\n",
           cur->properties->children->content);

#ifndef WIN32
     pdu->messageID = (int64_t)atoll((char *)
                                     cur->properties->children->content);
     DBG1("parseXmpMessage: message-ID is %lld\n",pdu->messageID);
#else
     pdu->messageID = (int64_t)_atoi64((char *)
                                       cur->properties->children->content);
     DBG1("parseXmpMessage: message-ID is %i64\n",pdu->messageID);
#endif

     /* get message type */
     if (cur->children == NULL) {
        fprintf(stderr,"parseXmpMessage: XMP message has no type\n");
        xmlFreeDoc(doc);
        return -1;
     }

     /* ignore blank/whitespace */
     cur = cur->children;
     while (xmlIsBlankNode(cur)) { cur = cur->next; }

     pdu->msgType = stringToMessageType((char *)cur->name);
     if (pdu->msgType == invalidMsgType) {
        fprintf(stderr,"parseXmpMessage: invalid XMP message type '%s'\n",
                cur->name);
        xmlFreeDoc(doc);
        return -1;
     }

     /* rest of message is a child of message-type tag */
     cur = cur->children;

     /* parse error-status tag/content (no attributes) */
     while (xmlIsBlankNode(cur)) { cur = cur->next; }
     if (cur == NULL) {
        fprintf(stderr,"No error-status\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->name,"error-status") != 0) {
        fprintf(stderr,"parseXmpMessage: no error-status tag '%s'\n",
                cur->name);
        xmlFreeDoc(doc);
        return -1;
     }
     /* the actual error-status is child */
     if (cur->children == NULL) {
        printf("No error-status content/value\n");
        xmlFreeDoc(doc);
        return -1;
     }
     pdu->errorStatus = stringToErrorStatus((char *)cur->children->content);

     /* parse error-index tag/content (no attributes) */
     cur = cur->next;
     while (xmlIsBlankNode(cur)) { cur = cur->next; }
     if (cur == NULL) {
        fprintf(stderr,"No error-index\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->name,"error-index") != 0) {
        fprintf(stderr,"parseXmpMessage: no error-index tag '%s'\n",
                cur->name);
        xmlFreeDoc(doc);
        return -1;
     }
     /* the actual error-index is child */
     if (cur->children == NULL) {
        fprintf(stderr,"No error-index content/value\n");
        xmlFreeDoc(doc);
        return -1;
     }
     pdu->errorIndex = atoi((char *)cur->children->content);

     /* if its a table operation, get table MIB, table name,
        keyvalue, maxrows */

     if ((pdu->msgType > SetRequest) && (pdu->msgType < Trap)) {
        cur = cur->next;
        while (xmlIsBlankNode(cur)) { cur = cur->next; }
        if (cur == NULL) {
           fprintf(stderr,"No tableInfo\n");
           xmlFreeDoc(doc);
           return -1;
        }

        if (parseXmpTableInfo(cur,pdu) < 0) {
   	   fprintf(stderr,"Failed to parse tableInfo\n");
           xmlFreeDoc(doc); 
           return -1;
        }

     } /* table operation -- Select, Insert, Delete, Update */

     DBG("parseXmpMessage: starting to parse variable list\n");

     /* parse variable list */
     cur = cur->next;
     while (xmlIsBlankNode(cur)) { cur = cur->next; }
     if (cur == NULL) {
        fprintf(stderr,"No variableList\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->name,"variableList") != 0) {
        fprintf(stderr,"parseXmpMessage: no variableList tag '%s'\n",
                cur->name);
        xmlFreeDoc(doc);
        return -1;
     }
     /* actual variables should be children of variableList */
     if (cur->children == NULL) {
        fprintf(stderr,"parseXmpMessage: no variables\n");
	xmlFreeDoc(doc);
        return -1;
     }

     cur = cur->children;
     while (xmlIsBlankNode(cur)) { cur = cur->next; }

     pdu->numberVariables = 0; 
     pdu->varList = NULL;
     pdu->varListTail = NULL;
     while (cur != NULL) {

       if (strcmp((char *)cur->name,"variable") != 0) {
	  fprintf(stderr,"parseXmpMessage: did not find variable '%s'\n",
                  cur->name);
          xmlFreeDoc(doc);
          return -1;
       }

       /* parse the variable */
       DBG("parseXmpMessage: calling parseXmpVariable\n");
       if (parseXmpVariable(cur,pdu) < 0) {
          xmlFreeDoc(doc);
          /* free allocated space in variable list? */
          fprintf(stderr,"parseXmpMessage: parse variable failed\n");
          xmpPduCleanup(pdu);
          return -1;
       }

       /* on to the next variable; skip past white-space */
       cur = cur->next;
       while (xmlIsBlankNode(cur)) { cur = cur->next; }

     }

     /* done parsing message */
     xmlFreeDoc(doc);

     /* should we call xmlCleanupParser -- NO! */

     if (xmp_session_verbose)
        printf("Done parsing XMP message, %d variables\n",
               pdu->numberVariables);

     return 1;
}

/* ************************************************************* */
/* given an internal-format XMP message, print it out */

int dumpInternalXmpMessage(xmpPdu_t *pdu)
{
     xmpVariable_t *varptr;

     if (pdu == NULL) {
        fprintf(stderr,"dumpInternalXmpMessage: invalid arg(s)\n");
        return -1;
     }

     printf("Message: type:%d, id=%lld\n",pdu->msgType,pdu->messageID);
     printf("errorStatus:%d errorIndex:%d\n",pdu->errorStatus,
            pdu->errorIndex);

     if ((pdu->msgType > SetRequest) && (pdu->msgType < Trap)) {
        printf("TableInfo: mib:%s table:%s key:'%s' maxRows:%d\n",
               pdu->tableMib,
               pdu->tableName,
               pdu->tableKey,
               pdu->maxRows);
     }

     printf("Variables: %d\n",pdu->numberVariables);
     
     varptr = pdu->varList;
     while (varptr != NULL) {

       if (varptr->valueStrLen > 4095) {
           fprintf(stderr,"dumpInternalXmpMessage: valueBuf is 0x%lx\n",
                   varptr->valueBuf);
           printf("\tVar Name:'%s,%s,%s' Syntax:%d Len:%d Value:'%s'\n",
                  varptr->varName.mibName,
                  varptr->varName.objName,
                  varptr->varName.key,
                  varptr->varSyntax,
                  varptr->valueStrLen,
                  varptr->valueBuf);

       }
       else {
          printf("\tVar Name:'%s,%s,%s' Syntax:%d Value:'%s'\n",  
                 varptr->varName.mibName,
                 varptr->varName.objName,
                 varptr->varName.key,
                 varptr->varSyntax,
                 varptr->valueStr);
       }
       
       varptr = varptr->next;
     }

     return 1;
}

/* ************************************************************* */
/* given an XML-encoded message, print it out in some structured
   manner; 

   The easiest thing to do is to first parse/print
   using one of libxml2's functions
 */

int printXmpMessage(char *buffer, int bufsize)
{
     xmlDocPtr doc;
     xmlChar *xmlbuff;
     int buffersize;

     if ((buffer == NULL) || (bufsize < 32)) {
        fprintf(stderr,"printXmpMessage: invalid args\n");
        return -1;
     }

     doc = xmlParseMemory(buffer,bufsize);

     if (doc != NULL) {
        xmlDocDumpFormatMemory(doc,&xmlbuff,&buffersize,1);
        printf((char *)xmlbuff);

        xmlFree(xmlbuff);
        xmlFreeDoc(doc);
     }

     return 1;
}

/* ************************************************************* */
/* convert to/from string/xmpMessagetype */

int messageTypeToString(xmpMessageType_t mt, char *str, int maxlen)
{
     if (str == NULL) {
        return -1;
     }

     switch (mt) {
       case Response:
	  strncpy(str,"Response",maxlen-1);
 	  break;
       case GetRequest:
	  strncpy(str,"GetRequest",maxlen-1);
 	  break;
       case SetRequest:
	  strncpy(str,"SetRequest",maxlen-1);
 	  break;

       case SelectTableRequest:
	  strncpy(str,"SelectTableRequest",maxlen-1);
 	  break;
       case InsertTableRequest:
	  strncpy(str,"InsertTableRequest",maxlen-1);
 	  break;
       case DeleteTableRequest:
	  strncpy(str,"DeleteTableRequest",maxlen-1);
 	  break;
       case UpdateTableRequest:
	  strncpy(str,"UpdateTableRequest",maxlen-1);
 	  break;

       case Trap:
	  strncpy(str,"Trap",maxlen-1);
 	  break;
       case Information:
	  strncpy(str,"Information",maxlen-1);
 	  break;

       default:
  	  return -1;
     }

     return strlen(str);
}

xmpMessageType_t stringToMessageType(char *str)
{
     if (str == NULL) {
        return invalidMsgType;
     }

     if (strcmp(str,"GetRequest") == 0) { return GetRequest; }
     if (strcmp(str,"Response") == 0) { return Response; }
     if (strcmp(str,"SetRequest") == 0) { return SetRequest; }

     if (strcmp(str,"SelectTableRequest") == 0) { return SelectTableRequest; }
     if (strcmp(str,"InsertTableRequest") == 0) { return InsertTableRequest; }
     if (strcmp(str,"DeleteTableRequest") == 0) { return DeleteTableRequest; }
     if (strcmp(str,"UpdateTableRequest") == 0) { return UpdateTableRequest; }

     if (strcmp(str,"Trap") == 0) { return Trap; }
     if (strcmp(str,"Information") == 0) { return Information; }

     return invalidMsgType;
}

/* ************************************************************* */

int syntaxToString(xmpSyntax_t syn, char *str, int maxlen)
{
     if (str == NULL) {
        return -1;
     }

     switch (syn) {
       case counter:
 	  strncpy(str,"counter",maxlen-1);
 	  break;
       case ipV4Address:
 	  strncpy(str,"ipV4Address",maxlen-1);
 	  break;
       case ipV6Address:
 	  strncpy(str,"ipV6Address",maxlen-1);
 	  break;
       case dateTime:
 	  strncpy(str,"dateTime",maxlen-1);
 	  break;
       case gauge:
 	  strncpy(str,"gauge",maxlen-1);
 	  break;
       case integer:
 	  strncpy(str,"integer",maxlen-1);
 	  break;
       case unsignedInteger:
 	  strncpy(str,"unsignedInteger",maxlen-1);
 	  break;
       case bool:
 	  strncpy(str,"bool",maxlen-1);
 	  break;
       case macAddress:
 	  strncpy(str,"macAddress",maxlen-1);
 	  break;
       case physAddress:
 	  strncpy(str,"physAddress",maxlen-1);
 	  break;
       case displayString:
 	  strncpy(str,"displayString",maxlen-1);
 	  break;
       case binaryString:
 	  strncpy(str,"binaryString",maxlen-1);
 	  break;
       case nullSyntax:
 	  strncpy(str,"nullSyntax",maxlen-1);
 	  break;
       case table:
 	  strncpy(str,"table",maxlen-1);
 	  break;
       case floatingPoint:
 	  strncpy(str,"floatingPoint",maxlen-1);
 	  break;
       case unsupportedVariable:
	  strncpy(str,"unsupportedVariable",maxlen-1);
 	  break;
       case extendedBoolean:
	  strncpy(str,"extendedBoolean",maxlen-1);
 	  break;
       default:
          return -1;
     }
  
     return 1;  
}

xmpSyntax_t stringToSyntax(char *str)
{
     if (str == NULL) {
        return invalidSyntax;
     }

     if (strcmp(str,"counter") == 0) { return counter; }
     if (strcmp(str,"ipV4Address") == 0) { return ipV4Address; }
     if (strcmp(str,"ipV6Address") == 0) { return ipV6Address; }
     if (strcmp(str,"dateTime") == 0) { return dateTime; }
     if (strcmp(str,"gauge") == 0) { return gauge; }
     if (strcmp(str,"integer") == 0) { return integer; }
     if (strcmp(str,"unsignedInteger") == 0) { return unsignedInteger; }
     if (strcmp(str,"bool") == 0) { return bool; }
     if (strcmp(str,"boolean") == 0) { return bool; }
     if (strcmp(str,"macAddress") == 0) { return macAddress; }
     if (strcmp(str,"physAddress") == 0) { return physAddress; }
     if (strcmp(str,"displayString") == 0) { return displayString; }
     if (strcmp(str,"binaryString") == 0) { return binaryString; }
     if (strcmp(str,"nullSyntax") == 0) { return nullSyntax; }
     if (strcmp(str,"table") == 0) { return table; }
     if (strcmp(str,"floatingPoint") == 0) { return floatingPoint; }
     if (strcmp(str,"unsupportedVariable") == 0) { return unsupportedVariable; }
     if (strcmp(str,"extendedBoolean") == 0) { return extendedBoolean; }

     return invalidSyntax;
}

/* ************************************************************* */

int errorStatusToString(xmpErrorStatus_t e, char *str, int maxlen)
{
     if (str == NULL) {
        return -1;
     }

     switch (e) {
       case noError:
 	  strncpy(str,"noError",maxlen-1);
          break;
       case noErrorMoreData:
 	  strncpy(str,"noErrorMoreData",maxlen-1);
          break;
       case tooBig:
	  strncpy(str,"tooBig",maxlen-1);
          break;
       case noSuchName:
	  strncpy(str,"noSuchName",maxlen-1);
          break;
       case badValue:
 	  strncpy(str,"badValue",maxlen-1);
          break;
       case readOnly:
	  strncpy(str,"readOnly",maxlen-1);
          break;
       case genError:
	  strncpy(str,"genError",maxlen-1);
          break;
       case noAccess:
   	  strncpy(str,"noAccess",maxlen-1);
          break;
       case notTabular:
   	  strncpy(str,"notTabular",maxlen-1);
          break;
       case wrongType:
	  strncpy(str,"wrongType",maxlen-1);
          break;
       case wrongLength:
	  strncpy(str,"wrongLength",maxlen-1);
          break;
       case wrongEncoding:
	  strncpy(str,"wrongEncoding",maxlen-1);
          break;
       case wrongValue:
	  strncpy(str,"wrongValue",maxlen-1);
          break;
       case noCreation:
	  strncpy(str,"noCreation",maxlen-1);
          break;
       case inconsistentValue:
	  strncpy(str,"inconsistentValue",maxlen-1);
          break;
       case resourceUnavailable:
	  strncpy(str,"resourceUnavailable",maxlen-1);
          break;
       case commitFailed:
	  strncpy(str,"commitFailed",maxlen-1);
          break;
       case undoFailed:
	  strncpy(str,"undoFailed",maxlen-1);
          break;
       case authorizationError:
	  strncpy(str,"authorizationError",maxlen-1);
          break;
       case notWritable:
	  strncpy(str,"notWritable",maxlen-1);
          break;
       case inconsistentName:
	  strncpy(str,"inconsistentName",maxlen-1);
          break;
       case parseError:
	  strncpy(str,"parseError",maxlen-1);
          break;
       default:
  	  return -1;
     }

     return strlen(str);
}

xmpErrorStatus_t stringToErrorStatus(char *str)
{
     if (str == NULL) {
        return invalidErrorType;
     }

     if (strcmp(str,"noError") == 0) { return noError; }
     if (strcmp(str,"noErrorMoreData") == 0) { return noErrorMoreData; }
     if (strcmp(str,"tooBig") == 0) { return tooBig; }
     if (strcmp(str,"noSuchName") == 0) { return noSuchName; }
     if (strcmp(str,"badValue") == 0) { return badValue; }
     if (strcmp(str,"readOnly") == 0) { return readOnly; }
     if (strcmp(str,"genError") == 0) { return genError; }
     if (strcmp(str,"noAccess") == 0) { return noAccess; }
     if (strcmp(str,"notTabular") == 0) { return notTabular; }
     if (strcmp(str,"wrongType") == 0) { return wrongType; }
     if (strcmp(str,"wrongLength") == 0) { return wrongLength; }
     if (strcmp(str,"wrongEncoding") == 0) { return wrongEncoding; }
     if (strcmp(str,"wrongValue") == 0) { return wrongValue; }
     if (strcmp(str,"noCreation") == 0) { return noCreation; }
     if (strcmp(str,"inconsistentValue") == 0) { return inconsistentValue; }
     if (strcmp(str,"resourceUnavailable") == 0) { return resourceUnavailable;}
     if (strcmp(str,"commitFailed") == 0) { return commitFailed; }
     if (strcmp(str,"undoFailed") == 0) { return undoFailed; }
     if (strcmp(str,"authorizationError") == 0) { return authorizationError; }
     if (strcmp(str,"notWritable") == 0) { return notWritable; }
     if (strcmp(str,"inconsistentName") == 0) { return inconsistentName; }
     if (strcmp(str,"parseError") == 0) { return parseError; }

     return invalidErrorType;
}

/* ************************************************************* */
