/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmpd_config.c */
/* rdk@krupczak.org */

/* 
   parse/build daemon's config file which also
   serves as backing store for config info;  Config-file
   is in XML format

*/

#ifdef WIN32
#include <windows.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#ifndef WIN32
#include <strings.h>
#include <sys/time.h>
#include <unistd.h>
#endif

#include <stdlib.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "xmp.h"
#include "xmp_session.h"
#include "globals.h"
#include "modules.h"
#include "xmp_users.h"

static time_t readConfigFileTime = 0;

/* ************************************************************* */
/* parse an xmpd_authenuser statement 

    <xmpd_authenuser>
       <user>%s</user>
       <protocol_ops>x,y,z</protocol_ops>
       <mibs>a,b,c</mibs>
    </xmpd_authenuser>

*/

static int parseAuthenUser(xmlNodePtr node)
{
     char *cptr;
     xmlNodePtr curNode;
     u_int protoOps;
     char authenUser[18];
     char protoOpsStr[512];
     char mibList[512];
     int i;

     if (node == NULL) {
        return -1;
     }

     if (strcmp((char *)node->name,"xmpd_authenuser") != 0) {
        fprintf(stderr,"parseAuthenUser: %s is not a user\n",node->name);
        return -1;
     }

     if (node->children == NULL) {
        fprintf(stderr,"authenuser does not have mib, object, key, etc.\n");
        return -1;
     }

     /* <user></user> */
     curNode = node->children;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"authenUser does not contain user . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"user") != 0) {
        fprintf(stderr,"'%s' is not user\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No user value\n");
        return -1;
     }
     strncpy(authenUser,(char *)curNode->children->content,16);
     authenUser[16] = 0;

     /* <protocol_ops></protocol_ops> */
     curNode = curNode->next;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"authenUser does not contain protocol_ops . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"protocol_ops") != 0) {
        fprintf(stderr,"'%s' is not protocol_ops\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No protocol_ops value\n");
        return -1;
     }
     strncpy(protoOpsStr,(char *)curNode->children->content,511);
     protoOpsStr[511] = 0;

     /* <mibs></mibs>*/
     curNode = curNode->next;
     while (xmlIsBlankNode(curNode)) { curNode = curNode->next; }
     if (curNode == NULL) {
        fprintf(stderr,"authenUser does not contain mibs . . .\n");
        return -1;
     }
     if (strcmp((char *)curNode->name,"mibs") != 0) {
        fprintf(stderr,"'%s' is not mibs\n",curNode->name);
        return -1;
     }
     if (curNode->children == NULL) {
        fprintf(stderr,"No mibs value\n");
        return -1;
     }
     strncpy(mibList,(char *)curNode->children->content,511);
     mibList[511] = 0;

     /* go through protoOpsStr digging out the comma-separate ops */
     cptr = protoOpsStr;
     protoOps = 0;
     for (i=0; i<strlen(protoOpsStr); i++) {
         protoOpsStr[i] = (char)tolower(protoOpsStr[i]);
     }

     if (strstr(protoOpsStr,"response") == 0)
        protoOps = protoOps | Response;
     if (strstr(protoOpsStr,"getrequest") == 0)
        protoOps = protoOps | GetRequest;
     if (strstr(protoOpsStr,"setrequest") == 0)
        protoOps = protoOps | SetRequest;
     if (strstr(protoOpsStr,"selecttablerequest") == 0)
        protoOps = protoOps | SelectTableRequest;
     if (strstr(protoOpsStr,"inserttablerequest") == 0)
        protoOps = protoOps | InsertTableRequest;
     if (strstr(protoOpsStr,"deletetablerequest") == 0)
        protoOps = protoOps | DeleteTableRequest;
     if (strstr(protoOpsStr,"updatetablerequest") == 0)
        protoOps = protoOps | UpdateTableRequest;
     if (strstr(protoOpsStr,"trap") == 0)
        protoOps = protoOps | Trap;
     if (strstr(protoOpsStr,"information") == 0)
        protoOps = protoOps | Information; 
     if (strstr(protoOpsStr,"all") == 0)
        protoOps = protoOps | XMP_ALL_OPS;

     return xmpAddUser(authenUser,protoOps,mibList);
}

/* ************************************************************* */
/* parse the XML tree for subsequent parsing/interpreting */

static int parseXmlConfigTree(xmlNodePtr ptr)
{
     if (xmpd_verbose)
        printf("parseXmlConfigTree: called with %lx\n",ptr);

     /* return on whitespace */
     if (xmlIsBlankNode(ptr)) {
        return 0;
     }

     if (ptr == NULL)
        return 0;
 
     if (strcmp((char *)ptr->name,"xmpd_port") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
	  if (xmpd_verbose)
  	     printf("Configfile: port '%s'\n",
                    ptr->children->content);
          xmpd_port = atoi((char *)ptr->children->content);
        }
     }

     if (strcmp((char *)ptr->name,"xmpd_trace") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
	   if (xmpd_verbose)
      	      printf("Configfile: trace '%s'\n",
                     ptr->children->content);
           xmpd_trace = atoi((char *)ptr->children->content);
        }
     }

     if (strcmp((char *)ptr->name,"xmpd_verbose") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
	  if (xmpd_verbose)
	     printf("Configfile: verbose '%s'\n",
                    ptr->children->content);
             xmpd_verbose = atoi((char *)ptr->children->content);
        }
     }
     if (strcmp((char *)ptr->name,"xmpd_daemon") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
           if (xmpd_verbose)
   	      printf("Configfile: daemon '%s'\n",
                     ptr->children->content);
           xmpd_daemon = atoi((char *)ptr->children->content);
        }
     }

     if (strcmp((char *)ptr->name,"xmpd_eventageoutinterval") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
           if (xmpd_verbose)
   	      printf("Configfile: event ageout interval '%s'\n",
                     ptr->children->content);
           eventAgeOutInterval = atoi((char *)ptr->children->content);
        }
     }

     if (strcmp((char *)ptr->name,"xmpd_defaultmanager") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
           if (xmpd_verbose)
   	      printf("Configfile: defaultmanager '%s'\n",
                     ptr->children->content);
           strncpy(defaultManager,ptr->children->content,255);
           defaultManager[255] = 0;

        }
     }

     if (strcmp((char *)ptr->name,"xmpd_autoupgrade") == 0) {
        if ((ptr->children != NULL) && (ptr->children->content != NULL)) {
           if (xmpd_verbose)
   	      printf("Configfile: xmpd_autoupgrade '%s'\n",
                     ptr->children->content);
           if (strstr(ptr->children->content,"true") != 0) {
	      autoUpgrade = 1;
           }
           if (strstr(ptr->children->content,"false") != 0) {
	      autoUpgrade = 0;
           }
        }
     }

     if (strcmp((char *)ptr->name,"xmpd_authenuser") == 0) {
        parseAuthenUser(ptr);
     }
     
     return 1;
}

/* ************************************************************* */
/* build the Xml tree for subsequent writing to file?
   the config-file may be simple enough not to have to go through
   this exercise 
*/
  
static int buildXmlConfigTree(xmlNodePtr ptr)
{
     return 1;
}

/* ************************************************************* */
/* write xmpd config-file either to configfilename
   or to homedir/xmpd.xml */

int writeXmpdConfigFile()
{
     FILE *fp;
     char timestr[32];
     time_t thetime;

     if (xmpd_config_file == NULL) {
        xmpd_config_file = (char *)malloc(strlen(xmpd_home_dir)+32);
        if (xmpd_config_file == NULL) {
 	   perror("malloc of space for config file name failed");
           return -1;
        }
#ifndef WIN32
        sprintf(xmpd_config_file,"%s/xmpd.xml",xmpd_home_dir);
#else
        sprintf(xmpd_config_file,"%s\\xmpd.xml",xmpd_home_dir);
#endif
     }

     /* open/create config-file */
     if ((fp = fopen(xmpd_config_file,"w")) == NULL) {
        perror("open/create of config file failed");
        xmpd_log("xmpd: failed to open config file");
        return -1;
     }

     fprintf(fp,"<?xml version=\"1.0\"?>\n\n");
     fprintf(fp,"<xmpd_config>\n\n");

     thetime = time(NULL);
#ifdef LINUX
     ctime_r(&thetime,timestr);
#endif
#ifdef SOLARIS
     ctime_r(&thetime,timestr,sizeof(timestr));
#endif
     if (timestr[strlen(timestr)-1] == '\n')
        timestr[strlen(timestr)-1] = 0;
     fprintf(fp,"<!-- Written by xmpd version %3.2f, %s -->\n\n",
             xmpd_version,timestr);

     fprintf(fp,"<xmpd_trace>%d</xmpd_trace>\n",xmpd_trace);

     fprintf(fp,"<xmpd_port>%d</xmpd_port>\n",xmpd_port);

     fprintf(fp,"<xmpd_verbose>%d</xmpd_verbose>\n",xmpd_verbose);

     fprintf(fp,"<xmpd_daemon>%d</xmpd_daemon>\n",xmpd_daemon);
     
     fprintf(fp,"<xmpd_eventageoutinterval>%d</xmpd_eventageoutinterval>\n",
	     eventAgeOutInterval);

     if (defaultManager[0] != 0) {
        fprintf(fp,"<xmpd_defaultmanager>%s</xmpd_defaultmanager>\n",
	        defaultManager);
     }

     if (autoUpgrade == 1) {
        fprintf(fp,"<xmpd_autoupgrade>true</xmpd_autoupgrade>\n\n");
     }
     else {
        fprintf(fp,"<xmpd_autoupgrade>false</xmpd_autoupgrade>\n\n");
     }

     fprintf(fp,"</xmpd_config>\n");
     fclose(fp);

     return 1;
}

/* ************************************************************* */
/* read/parse the config file looking for parameters, etc. */

int readXmpdConfigFile(char *configfilename) 
{
     char aPath[FILENAME_MAX+1];
#ifndef WIN32
     struct stat st;
#else
     struct _stat st;
#endif
     xmlDocPtr doc;
     xmlNodePtr cur;
     int newVersion;
     char version[32];
     
     if (configfilename == NULL) {
#ifndef WIN32
        sprintf(aPath,"%s/%s",xmpd_home_dir,"xmpd.xml");
#else
        sprintf(aPath,"%s\\%s",xmpd_home_dir,"xmpd.xml");
#endif
     }
     else {
        sprintf(aPath,"%s",configfilename);
     }

     if (xmpd_verbose)
        printf("readXmpdConfigFile: checking %s\n",aPath);
     DBG1("readXmpdConfigFile: checking %s\n",aPath);

     /* stat it to see if it exists */
#ifndef WIN32
     if (stat(aPath,&st) < 0) {
#else
     if (_stat(aPath,&st) < 0) {
#endif
        perror("stat of config file failed");
        xmpd_log("readXmpdConfigFile: failed to stat config file");
        return -1;
     }

#ifdef WIN32
#define	_S_ISTYPE(mode, mask)	(((mode) & _S_IFMT) == (mask))
#define	S_ISREG(mode)	 _S_ISTYPE((mode), _S_IFREG)
#endif

     if (! S_ISREG(st.st_mode)) {
        fprintf(stderr,"config file is not regular!\n");
        return -1;
     }

     if (st.st_mtime < readConfigFileTime) {
        return 0;
     }

     /* parse it! */
     if ((doc = xmlParseFile(aPath)) == NULL) {
        fprintf(stderr,"failed to xml parse config file\n");
        return -1;
     }
     if ((cur = xmlDocGetRootElement(doc)) == NULL) {
        fprintf(stderr,"empty config document\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->name,"xmpd_config") != 0) {
        fprintf(stderr,"document is not an xmpd config message, %s\n",
                cur->name);
        xmlFreeDoc(doc);
        return -1;
     }
     /* get version as attribute of xmpd_config */
     if ((cur->properties == NULL) || (cur->properties->children == NULL)) {
        fprintf(stderr,"config file has no version\n");
        xmlFreeDoc(doc);
        return -1;
     }
     if (strcmp((char *)cur->properties->name,"version") != 0) {
        fprintf(stderr,"config file: no version property\n");
        xmlFreeDoc(doc);
        return -1;
     }
     newVersion = atoi((char *)cur->properties->children->content);
     if (newVersion <= xmpd_config_version) {
        fprintf(stderr,"No need to re-read xmpd config file\n");
        xmlFreeDoc(doc);
        readConfigFileTime = time(NULL);
        return 0;
     }

     /* each config-statement should be a child of the root element */
     if (cur->children == NULL) {
        fprintf(stderr,"config file has no config statements\n");
        xmlFreeDoc(doc);
        return -1;
     }

     cur = cur->children;
     while (cur != NULL) {
       parseXmlConfigTree(cur);
       cur = cur->next;
     }

     xmlFreeDoc(doc);

     readConfigFileTime = time(NULL);
     xmpd_config_version = newVersion;

     /* add the module */
     sprintf(version,"%d",xmpd_config_version);
     addModule("xmpd.xml",version,"xmpd",NULL,dataModule);

     return 1;     
}

/* ************************************************************* */
